# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


from time import sleep
import typing

from PySide6.QtCore import Signal, Qt, QThreadPool
from PySide6.QtGui import QStandardItemModel, QStandardItem
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QListView,
)

from src.core.library import Library
from src.qt.helpers.function_iterator import FunctionIterator
from src.qt.helpers.custom_runnable import CustomRunnable
from src.qt.widgets.progress import ProgressWidget

# Only import for type checking/autocompletion, will not be imported at runtime.
if typing.TYPE_CHECKING:
    from src.qt.ts_qt import QtDriver


class MirrorEntriesModal(QWidget):
    done = Signal()

    def __init__(self, library: "Library", driver: "QtDriver"):
        super().__init__()
        self.lib = library
        self.driver = driver
        self.setWindowTitle(f"Mirror Entries")
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setMinimumSize(500, 400)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 6, 6, 6)

        self.desc_widget = QLabel()
        self.desc_widget.setObjectName("descriptionLabel")
        self.desc_widget.setWordWrap(True)
        self.desc_widget.setText(f"""
		Are you sure you want to mirror the following {len(self.lib.dupe_files)} Entries?
		""")
        self.desc_widget.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.list_view = QListView()
        self.model = QStandardItemModel()
        self.list_view.setModel(self.model)

        self.button_container = QWidget()
        self.button_layout = QHBoxLayout(self.button_container)
        self.button_layout.setContentsMargins(6, 6, 6, 6)
        self.button_layout.addStretch(1)

        self.cancel_button = QPushButton()
        self.cancel_button.setText("&Cancel")
        self.cancel_button.setDefault(True)
        self.cancel_button.clicked.connect(self.hide)
        self.button_layout.addWidget(self.cancel_button)

        self.mirror_button = QPushButton()
        self.mirror_button.setText("&Mirror")
        self.mirror_button.clicked.connect(self.hide)
        self.mirror_button.clicked.connect(lambda: self.mirror_entries())
        self.button_layout.addWidget(self.mirror_button)

        self.root_layout.addWidget(self.desc_widget)
        self.root_layout.addWidget(self.list_view)
        self.root_layout.addWidget(self.button_container)

    def refresh_list(self):
        self.desc_widget.setText(f"""
		Are you sure you want to mirror the following {len(self.lib.dupe_files)} Entries?
		""")

        self.model.clear()
        for i in self.lib.dupe_files:
            self.model.appendRow(QStandardItem(str(i)))

    def mirror_entries(self):
        # pb = QProgressDialog('', None, 0, len(self.lib.dupe_files))
        # # pb.setMaximum(len(self.lib.missing_files))
        # pb.setFixedSize(432, 112)
        # pb.setWindowFlags(pb.windowFlags() & ~Qt.WindowType.WindowCloseButtonHint)
        # pb.setWindowTitle('Mirroring Entries')
        # pb.setWindowModality(Qt.WindowModality.ApplicationModal)
        # pb.show()

        # r = CustomRunnable(lambda: self.mirror_entries_runnable(pb))
        # r.done.connect(lambda: self.done.emit())
        # r.done.connect(lambda: self.driver.preview_panel.refresh())
        # # r.done.connect(lambda: self.model.clear())
        # # QThreadPool.globalInstance().start(r)
        # r.run()

        iterator = FunctionIterator(self.mirror_entries_runnable)
        pw = ProgressWidget(
            window_title="Mirroring Entries",
            label_text=f"Mirroring 1/{len(self.lib.dupe_files)} Entries...",
            cancel_button_text=None,
            minimum=0,
            maximum=len(self.lib.dupe_files),
        )
        pw.show()
        iterator.value.connect(lambda x: pw.update_progress(x + 1))
        iterator.value.connect(
            lambda x: pw.update_label(
                f"Mirroring {x+1}/{len(self.lib.dupe_files)} Entries..."
            )
        )
        r = CustomRunnable(lambda: iterator.run())
        QThreadPool.globalInstance().start(r)
        r.done.connect(
            lambda: (
                pw.hide(),
                pw.deleteLater(),
                self.driver.preview_panel.update_widgets(),
                self.done.emit(),
            )
        )

    def mirror_entries_runnable(self):
        mirrored: list = []
        for i, dupe in enumerate(self.lib.dupe_files):
            # pb.setValue(i)
            # pb.setLabelText(f'Mirroring {i}/{len(self.lib.dupe_files)} Entries')
            entry_id_1 = self.lib.get_entry_id_from_filepath(dupe[0])
            entry_id_2 = self.lib.get_entry_id_from_filepath(dupe[1])
            self.lib.mirror_entry_fields([entry_id_1, entry_id_2])
            sleep(0.005)
            yield i
        for d in mirrored:
            self.lib.dupe_files.remove(d)
        # self.driver.filter_items('')
        # self.done.emit()



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


import os
import typing

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QFileDialog,
)

from src.core.library import Library
from src.qt.modals.mirror_entities import MirrorEntriesModal

# Only import for type checking/autocompletion, will not be imported at runtime.
if typing.TYPE_CHECKING:
    from src.qt.ts_qt import QtDriver


class FixDupeFilesModal(QWidget):
    # done = Signal(int)
    def __init__(self, library: "Library", driver: "QtDriver"):
        super().__init__()
        self.lib = library
        self.driver = driver
        self.count = -1
        self.filename = ""
        self.setWindowTitle(f"Fix Duplicate Files")
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setMinimumSize(400, 300)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 6, 6, 6)

        self.desc_widget = QLabel()
        self.desc_widget.setObjectName("descriptionLabel")
        self.desc_widget.setWordWrap(True)
        self.desc_widget.setStyleSheet(
            # 'background:blue;'
            "text-align:left;"
            # 'font-weight:bold;'
            # 'font-size:14px;'
            # 'padding-top: 6px'
            ""
        )
        self.desc_widget.setText(
            """TagStudio supports importing DupeGuru results to manage duplicate files."""
        )
        self.desc_widget.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.dupe_count = QLabel()
        self.dupe_count.setObjectName("dupeCountLabel")
        self.dupe_count.setStyleSheet(
            # 'background:blue;'
            # 'text-align:center;'
            "font-weight:bold;"
            "font-size:14px;"
            # 'padding-top: 6px'
            ""
        )
        self.dupe_count.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.file_label = QLabel()
        self.file_label.setObjectName("fileLabel")
        # self.file_label.setStyleSheet(
        # 								# 'background:blue;'
        # 						 		# 'text-align:center;'
        # 								'font-weight:bold;'
        # 								'font-size:14px;'
        # 								# 'padding-top: 6px'
        # 								'')
        # self.file_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.file_label.setText("No DupeGuru File Selected")

        self.open_button = QPushButton()
        self.open_button.setText("&Load DupeGuru File")
        self.open_button.clicked.connect(lambda: self.select_file())

        self.mirror_button = QPushButton()
        self.mirror_modal = MirrorEntriesModal(self.lib, self.driver)
        self.mirror_modal.done.connect(lambda: self.refresh_dupes())
        self.mirror_button.setText("&Mirror Entries")
        self.mirror_button.clicked.connect(lambda: self.mirror_modal.show())
        self.mirror_desc = QLabel()
        self.mirror_desc.setWordWrap(True)
        self.mirror_desc.setText(
            """Mirror the Entry data across each duplicate match set, combining all data while not removing or duplicating fields. This operation will not delete any files or data."""
        )

        # self.mirror_delete_button = QPushButton()
        # self.mirror_delete_button.setText('Mirror && Delete')

        self.advice_label = QLabel()
        self.advice_label.setWordWrap(True)
        self.advice_label.setText(
            """After mirroring, you're free to use DupeGuru to delete the unwanted files. Afterwards, use TagStudio's "Fix Unlinked Entries" feature in the Tools menu in order to delete the unlinked Entries."""
        )

        self.button_container = QWidget()
        self.button_layout = QHBoxLayout(self.button_container)
        self.button_layout.setContentsMargins(6, 6, 6, 6)
        self.button_layout.addStretch(1)

        self.done_button = QPushButton()
        self.done_button.setText("&Done")
        # self.save_button.setAutoDefault(True)
        self.done_button.setDefault(True)
        self.done_button.clicked.connect(self.hide)
        # self.done_button.clicked.connect(lambda: self.done.emit(self.combo_box.currentIndex()))
        # self.save_button.clicked.connect(lambda: save_callback(widget.get_content()))
        self.button_layout.addWidget(self.done_button)

        # self.returnPressed.connect(lambda: self.done.emit(self.combo_box.currentIndex()))

        # self.done.connect(lambda x: callback(x))

        self.root_layout.addWidget(self.desc_widget)
        self.root_layout.addWidget(self.dupe_count)
        self.root_layout.addWidget(self.file_label)
        self.root_layout.addWidget(self.open_button)
        # self.mirror_delete_button.setHidden(True)

        self.root_layout.addWidget(self.mirror_button)
        self.root_layout.addWidget(self.mirror_desc)
        # self.root_layout.addWidget(self.mirror_delete_button)
        self.root_layout.addWidget(self.advice_label)
        # self.root_layout.setStretch(1,2)
        self.root_layout.addStretch(1)
        self.root_layout.addWidget(self.button_container)

        self.set_dupe_count(self.count)

    def select_file(self):
        qfd = QFileDialog(self, "Open DupeGuru Results File", str(self.lib.library_dir))
        qfd.setFileMode(QFileDialog.FileMode.ExistingFile)
        qfd.setNameFilter("DupeGuru Files (*.dupeguru)")
        if qfd.exec_():
            filename = qfd.selectedFiles()
            if filename:
                self.set_filename(filename[0])

    def set_filename(self, filename: str):
        if filename:
            self.file_label.setText(filename)
        else:
            self.file_label.setText("No DupeGuru File Selected")
        self.filename = filename
        self.refresh_dupes()
        self.mirror_modal.refresh_list()

    def refresh_dupes(self):
        self.lib.refresh_dupe_files(self.filename)
        self.set_dupe_count(len(self.lib.dupe_files))

    def set_dupe_count(self, count: int):
        self.count = count
        if self.count < 0:
            self.mirror_button.setDisabled(True)
            self.dupe_count.setText(f"Duplicate File Matches: N/A")
        elif self.count == 0:
            self.mirror_button.setDisabled(True)
            self.dupe_count.setText(f"Duplicate File Matches: {count}")
        else:
            self.mirror_button.setDisabled(False)
            self.dupe_count.setText(f"Duplicate File Matches: {count}")



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

import typing

from PySide6.QtCore import QObject, Signal, QThreadPool

from src.core.library import Library
from src.qt.helpers.function_iterator import FunctionIterator
from src.qt.helpers.custom_runnable import CustomRunnable
from src.qt.widgets.progress import ProgressWidget

# Only import for type checking/autocompletion, will not be imported at runtime.
if typing.TYPE_CHECKING:
    from src.qt.ts_qt import QtDriver


class RelinkUnlinkedEntries(QObject):
    done = Signal()

    def __init__(self, library: "Library", driver: "QtDriver"):
        super().__init__()
        self.lib = library
        self.driver = driver
        self.fixed = 0

    def repair_entries(self):
        iterator = FunctionIterator(self.lib.fix_missing_files)

        pw = ProgressWidget(
            window_title="Relinking Entries",
            label_text="",
            cancel_button_text=None,
            minimum=0,
            maximum=len(self.lib.missing_files),
        )

        pw.show()

        iterator.value.connect(lambda x: pw.update_progress(x[0] + 1))
        iterator.value.connect(
            lambda x: (
                self.increment_fixed() if x[1] else (),
                pw.update_label(
                    f"Attempting to Relink {x[0]+1}/{len(self.lib.missing_files)} Entries, {self.fixed} Successfully Relinked"
                ),
            )
        )

        r = CustomRunnable(lambda: iterator.run())
        r.done.connect(
            lambda: (pw.hide(), pw.deleteLater(), self.done.emit(), self.reset_fixed())
        )
        QThreadPool.globalInstance().start(r)

    def increment_fixed(self):
        self.fixed += 1

    def reset_fixed(self):
        self.fixed = 0



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

import typing

from PySide6.QtCore import Signal, Qt, QThreadPool
from PySide6.QtGui import QStandardItemModel, QStandardItem
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QListView,
)

from src.core.library import ItemType, Library
from src.qt.helpers.custom_runnable import CustomRunnable
from src.qt.helpers.function_iterator import FunctionIterator
from src.qt.widgets.progress import ProgressWidget

# Only import for type checking/autocompletion, will not be imported at runtime.
if typing.TYPE_CHECKING:
    from src.qt.ts_qt import QtDriver


class DeleteUnlinkedEntriesModal(QWidget):
    done = Signal()

    def __init__(self, library: "Library", driver: "QtDriver"):
        super().__init__()
        self.lib = library
        self.driver = driver
        self.setWindowTitle("Delete Unlinked Entries")
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setMinimumSize(500, 400)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 6, 6, 6)

        self.desc_widget = QLabel()
        self.desc_widget.setObjectName("descriptionLabel")
        self.desc_widget.setWordWrap(True)
        self.desc_widget.setText(f"""
		Are you sure you want to delete the following {len(self.lib.missing_files)} entries?
		""")
        self.desc_widget.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.list_view = QListView()
        self.model = QStandardItemModel()
        self.list_view.setModel(self.model)

        self.button_container = QWidget()
        self.button_layout = QHBoxLayout(self.button_container)
        self.button_layout.setContentsMargins(6, 6, 6, 6)
        self.button_layout.addStretch(1)

        self.cancel_button = QPushButton()
        self.cancel_button.setText("&Cancel")
        self.cancel_button.setDefault(True)
        self.cancel_button.clicked.connect(self.hide)
        self.button_layout.addWidget(self.cancel_button)

        self.delete_button = QPushButton()
        self.delete_button.setText("&Delete")
        self.delete_button.clicked.connect(self.hide)
        self.delete_button.clicked.connect(lambda: self.delete_entries())
        self.button_layout.addWidget(self.delete_button)

        self.root_layout.addWidget(self.desc_widget)
        self.root_layout.addWidget(self.list_view)
        self.root_layout.addWidget(self.button_container)

    def refresh_list(self):
        self.desc_widget.setText(f"""
		Are you sure you want to delete the following {len(self.lib.missing_files)} entries?
		""")

        self.model.clear()
        for i in self.lib.missing_files:
            self.model.appendRow(QStandardItem(str(i)))

    def delete_entries(self):
        iterator = FunctionIterator(self.lib.remove_missing_files)

        pw = ProgressWidget(
            window_title="Deleting Entries",
            label_text="",
            cancel_button_text=None,
            minimum=0,
            maximum=len(self.lib.missing_files),
        )
        pw.show()

        iterator.value.connect(lambda x: pw.update_progress(x[0] + 1))
        iterator.value.connect(
            lambda x: pw.update_label(
                f"Deleting {x[0]+1}/{len(self.lib.missing_files)} Unlinked Entries"
            )
        )
        iterator.value.connect(
            lambda x: self.driver.purge_item_from_navigation(ItemType.ENTRY, x[1])
        )

        r = CustomRunnable(lambda: iterator.run())
        QThreadPool.globalInstance().start(r)
        r.done.connect(lambda: (pw.hide(), pw.deleteLater(), self.done.emit()))



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


import math
import os
from types import FunctionType
from pathlib import Path

from PIL import Image
from PySide6.QtCore import Signal, Qt, QEvent
from PySide6.QtGui import QEnterEvent, QAction
from PySide6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton

from src.core.library import Library, Tag
from src.core.palette import ColorType, get_tag_color


ERROR = f"[ERROR]"
WARNING = f"[WARNING]"
INFO = f"[INFO]"


class TagWidget(QWidget):
    edit_icon_128: Image.Image = Image.open(
        str(Path(__file__).parents[3] / "resources/qt/images/edit_icon_128.png")
    ).resize((math.floor(14 * 1.25), math.floor(14 * 1.25)))
    edit_icon_128.load()
    on_remove = Signal()
    on_click = Signal()
    on_edit = Signal()

    def __init__(
        self,
        library: Library,
        tag: Tag,
        has_edit: bool,
        has_remove: bool,
        on_remove_callback: FunctionType = None,
        on_click_callback: FunctionType = None,
        on_edit_callback: FunctionType = None,
    ) -> None:
        super().__init__()
        self.lib = library
        self.tag = tag
        self.has_edit: bool = has_edit
        self.has_remove: bool = has_remove
        # self.bg_label = QLabel()
        # self.setStyleSheet('background-color:blue;')

        # if on_click_callback:
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.base_layout = QVBoxLayout(self)
        self.base_layout.setObjectName("baseLayout")
        self.base_layout.setContentsMargins(0, 0, 0, 0)

        self.bg_button = QPushButton(self)
        self.bg_button.setFlat(True)
        self.bg_button.setText(tag.display_name(self.lib).replace("&", "&&"))
        if has_edit:
            edit_action = QAction("Edit", self)
            edit_action.triggered.connect(on_edit_callback)
            edit_action.triggered.connect(self.on_edit.emit)
            self.bg_button.addAction(edit_action)
        # if on_click_callback:
        self.bg_button.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)
        # if has_remove:
        # 	remove_action = QAction('Remove', self)
        # 	# remove_action.triggered.connect(on_remove_callback)
        # 	remove_action.triggered.connect(self.on_remove.emit())
        # 	self.bg_button.addAction(remove_action)
        search_for_tag_action = QAction("Search for Tag", self)
        # search_for_tag_action.triggered.connect(on_click_callback)
        search_for_tag_action.triggered.connect(self.on_click.emit)
        self.bg_button.addAction(search_for_tag_action)
        add_to_search_action = QAction("Add to Search", self)
        self.bg_button.addAction(add_to_search_action)

        self.inner_layout = QHBoxLayout()
        self.inner_layout.setObjectName("innerLayout")
        self.inner_layout.setContentsMargins(2, 2, 2, 2)
        # self.inner_layout.setAlignment(Qt.AlignmentFlag.AlignRight)

        # self.inner_container = QWidget()
        # self.inner_container.setLayout(self.inner_layout)
        # self.base_layout.addWidget(self.inner_container)
        self.bg_button.setLayout(self.inner_layout)
        self.bg_button.setMinimumSize(math.ceil(22 * 1.5), 22)

        # self.bg_button.setStyleSheet(
        # 	f'QPushButton {{'
        # 	f'border: 2px solid #8f8f91;'
        # 	f'border-radius: 6px;'
        # 	f'background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,stop: 0 {ColorType.PRIMARY}, stop: 1 {ColorType.BORDER});'
        # 	f'min-width: 80px;}}')

        self.bg_button.setStyleSheet(
            # f'background: {get_tag_color(ColorType.PRIMARY, tag.color)};'
            f"QPushButton{{"
            f"background: {get_tag_color(ColorType.PRIMARY, tag.color)};"
            # f'background-color: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 {get_tag_color(ColorType.PRIMARY, tag.color)}, stop:1.0 {get_tag_color(ColorType.BORDER, tag.color)});'
            # f"border-color:{get_tag_color(ColorType.PRIMARY, tag.color)};"
            f"color: {get_tag_color(ColorType.TEXT, tag.color)};"
            f"font-weight: 600;"
            f"border-color:{get_tag_color(ColorType.BORDER, tag.color)};"
            f"border-radius: 6px;"
            f"border-style:solid;"
            f"border-width: {math.ceil(1*self.devicePixelRatio())}px;"
            # f'border-top:2px solid {get_tag_color(ColorType.LIGHT_ACCENT, tag.color)};'
            # f'border-bottom:2px solid {get_tag_color(ColorType.BORDER, tag.color)};'
            # f'border-left:2px solid {get_tag_color(ColorType.BORDER, tag.color)};'
            # f'border-right:2px solid {get_tag_color(ColorType.BORDER, tag.color)};'
            # f'padding-top: 0.5px;'
            f"padding-right: 4px;"
            f"padding-bottom: 1px;"
            f"padding-left: 4px;"
            f"font-size: 13px"
            f"}}"
            f"QPushButton::hover{{"
            # f'background: {get_tag_color(ColorType.LIGHT_ACCENT, tag.color)};'
            # f'background-color: qlineargradient(x1:0, y1:0, x2:0, y2:1, stop:0 {get_tag_color(ColorType.PRIMARY, tag.color)}, stop:1.0 {get_tag_color(ColorType.BORDER, tag.color)});'
            # f"border-color:{get_tag_color(ColorType.PRIMARY, tag.color)};"
            # f"color: {get_tag_color(ColorType.TEXT, tag.color)};"
            f"border-color:{get_tag_color(ColorType.LIGHT_ACCENT, tag.color)};"
            f"}}"
        )

        # self.renderer = ThumbRenderer()
        # self.renderer.updated.connect(lambda ts, i, s, ext: (self.update_thumb(ts, image=i),
        # 													 self.update_size(
        # 														 ts, size=s),
        # 													 self.set_extension(ext)))

        # self.bg_button.setLayout(self.base_layout)

        self.base_layout.addWidget(self.bg_button)
        # self.setMinimumSize(self.bg_button.size())

        # logging.info(tag.color)
        if has_remove:
            self.remove_button = QPushButton(self)
            self.remove_button.setFlat(True)
            self.remove_button.setText("–")
            self.remove_button.setHidden(True)
            self.remove_button.setStyleSheet(
                f"color: {get_tag_color(ColorType.PRIMARY, tag.color)};"
                f"background: {get_tag_color(ColorType.TEXT, tag.color)};"
                # f"color: {'black' if color not in ['black', 'gray', 'dark gray', 'cool gray', 'warm gray', 'blue', 'purple', 'violet'] else 'white'};"
                # f"border-color: {get_tag_color(ColorType.BORDER, tag.color)};"
                f"font-weight: 800;"
                # f"border-color:{'black' if color not in [
                # 'black', 'gray', 'dark gray',
                # 'cool gray', 'warm gray', 'blue',
                # 'purple', 'violet'] else 'white'};"
                f"border-radius: 4px;"
                # f'border-style:solid;'
                f"border-width:0;"
                # f'padding-top: 1.5px;'
                # f'padding-right: 4px;'
                f"padding-bottom: 4px;"
                # f'padding-left: 4px;'
                f"font-size: 14px"
            )
            self.remove_button.setMinimumSize(19, 19)
            self.remove_button.setMaximumSize(19, 19)
            # self.remove_button.clicked.connect(on_remove_callback)
            self.remove_button.clicked.connect(self.on_remove.emit)

        # NOTE: No more edit button! Just make it a right-click option.
        # self.edit_button = QPushButton(self)
        # self.edit_button.setFlat(True)
        # self.edit_button.setText('Edit')
        # self.edit_button.setIcon(QPixmap.fromImage(ImageQt.ImageQt(self.edit_icon_128)))
        # self.edit_button.setIconSize(QSize(14,14))
        # self.edit_button.setHidden(True)
        # self.edit_button.setStyleSheet(f'color: {color};'
        # 						    f"background: {'black' if color not in ['black', 'gray', 'dark gray', 'cool gray', 'warm gray', 'blue', 'purple', 'violet'] else 'white'};"
        # 							# f"color: {'black' if color not in ['black', 'gray', 'dark gray', 'cool gray', 'warm gray', 'blue', 'purple', 'violet'] else 'white'};"
        # 							f"border-color: {'black' if color not in ['black', 'gray', 'dark gray', 'cool gray', 'warm gray', 'blue', 'purple', 'violet'] else 'white'};"
        # 							f'font-weight: 600;'
        # 							# f"border-color:{'black' if color not in [
        # 							# 'black', 'gray', 'dark gray',
        # 							# 'cool gray', 'warm gray', 'blue',
        # 							# 'purple', 'violet'] else 'white'};"
        # 							# f'QPushButton{{border-image: url(:/images/edit_icon_128.png);}}'
        # 							# f'QPushButton{{border-image: url(:/images/edit_icon_128.png);}}'
        # 							f'border-radius: 4px;'
        # 							# f'border-style:solid;'
        # 							# f'border-width:1px;'
        # 							f'padding-top: 1.5px;'
        # 							f'padding-right: 4px;'
        # 							f'padding-bottom: 3px;'
        # 							f'padding-left: 4px;'
        # 							f'font-size: 14px')
        # self.edit_button.setMinimumSize(18,18)
        # # self.edit_button.setMaximumSize(18,18)

        # self.inner_layout.addWidget(self.edit_button)
        if has_remove:
            self.inner_layout.addWidget(self.remove_button)
        self.inner_layout.addStretch(1)

        # NOTE: Do this if you don't want the tag to stretch, like in a search.
        # self.bg_button.setMaximumWidth(self.bg_button.sizeHint().width())

        # self.set_click(on_click_callback)
        self.bg_button.clicked.connect(self.on_click.emit)

        # self.setMinimumSize(50,20)

    # def set_name(self, name:str):
    # 	self.bg_label.setText(str)

    # def on_remove(self):
    # 	if self.item and self.item[0] == ItemType.ENTRY:
    # 		if self.field_index >= 0:
    # 			self.lib.get_entry(self.item[1]).remove_tag(self.tag.id, self.field_index)
    # 		else:
    # 			self.lib.get_entry(self.item[1]).remove_tag(self.tag.id)

    # def set_click(self, callback):
    # 	try:
    # 		self.bg_button.clicked.disconnect()
    # 	except RuntimeError:
    # 		pass
    # 	if callback:
    # 		self.bg_button.clicked.connect(callback)

    # def set_click(self, function):
    # 	try:
    # 		self.bg.clicked.disconnect()
    # 	except RuntimeError:
    # 		pass
    # 	# self.bg.clicked.connect(lambda checked=False, filepath=filepath: open_file(filepath))
    # 	# self.bg.clicked.connect(function)

    def enterEvent(self, event: QEnterEvent) -> None:
        if self.has_remove:
            self.remove_button.setHidden(False)
        # self.edit_button.setHidden(False)
        self.update()
        return super().enterEvent(event)

    def leaveEvent(self, event: QEvent) -> None:
        if self.has_remove:
            self.remove_button.setHidden(True)
        # self.edit_button.setHidden(True)
        self.update()
        return super().leaveEvent(event)



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


from PIL import Image
from PySide6.QtCore import Qt
from PySide6.QtGui import QGuiApplication
from src.qt.helpers.gradient import linear_gradient

# TODO: Consolidate the built-in QT theme values with the values
# here, in enums.py, and in palette.py.
_THEME_DARK_FG: str = "#FFFFFF55"
_THEME_LIGHT_FG: str = "#000000DD"


def theme_fg_overlay(image: Image.Image) -> Image.Image:
    """
    Overlay the foreground theme color onto an image.

    Args:
        image (Image): The PIL Image object to apply an overlay to.
    """

    overlay_color = (
        _THEME_DARK_FG
        if QGuiApplication.styleHints().colorScheme() is Qt.ColorScheme.Dark
        else _THEME_LIGHT_FG
    )
    im = Image.new(mode="RGBA", size=image.size, color=overlay_color)
    return _apply_overlay(image, im)


def gradient_overlay(image: Image.Image, gradient=list[str]) -> Image.Image:
    """
    Overlay a color gradient onto an image.

    Args:
        image (Image): The PIL Image object to apply an overlay to.
        gradient (list[str): A list of string hex color codes for use as
            the colors of the gradient.
    """

    im: Image.Image = _apply_overlay(image, linear_gradient(image.size, gradient))
    return im


def _apply_overlay(image: Image.Image, overlay: Image.Image) -> Image.Image:
    """
    Internal method to apply an overlay on top of an image, using
    the image's alpha channel as a mask.

    Args:
        image (Image): The PIL Image object to apply an overlay to.
        overlay (Image): The PIL Image object to act as the overlay contents.
    """
    im: Image.Image = Image.new(mode="RGBA", size=image.size, color="#00000000")
    im.paste(overlay, (0, 0), mask=image)
    return im



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


from typing import Optional

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout, QProgressDialog


class ProgressWidget(QWidget):
    """Prebuilt thread-safe progress bar widget."""

    def __init__(
        self,
        window_title: str,
        label_text: str,
        cancel_button_text: Optional[str],
        minimum: int,
        maximum: int,
    ):
        super().__init__()
        self.root = QVBoxLayout(self)
        self.pb = QProgressDialog(
            labelText=label_text,
            minimum=minimum,
            cancelButtonText=cancel_button_text,
            maximum=maximum,
        )
        self.root.addWidget(self.pb)
        self.setFixedSize(432, 112)
        self.setWindowFlags(
            self.pb.windowFlags() & ~Qt.WindowType.WindowCloseButtonHint
        )
        self.setWindowTitle(window_title)
        self.setWindowModality(Qt.WindowModality.ApplicationModal)

    def update_label(self, text: str):
        self.pb.setLabelText(text)

    def update_progress(self, value: int):
        self.pb.setValue(value)



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

import logging
import os
import subprocess
import shutil
import sys
import traceback
from pathlib import Path

from PySide6.QtWidgets import QLabel
from PySide6.QtCore import Qt

ERROR = f"[ERROR]"
WARNING = f"[WARNING]"
INFO = f"[INFO]"

logging.basicConfig(format="%(message)s", level=logging.INFO)


def open_file(path: str | Path, file_manager: bool = False):
    """Open a file in the default application or file explorer.

    Args:
            path (str): The path to the file to open.
            file_manager (bool, optional): Whether to open the file in the file manager (e.g. Finder on macOS).
                    Defaults to False.
    """
    _path = str(path)
    logging.info(f"Opening file: {_path}")
    if not os.path.exists(_path):
        logging.error(f"File not found: {_path}")
        return
    try:
        if sys.platform == "win32":
            normpath = os.path.normpath(_path)
            if file_manager:
                command_name = "explorer"
                command_args = '/select,"' + normpath + '"'
                # For some reason, if the args are passed in a list, this will error when the path has spaces, even while surrounded in double quotes
                subprocess.Popen(
                    command_name + command_args,
                    shell=True,
                    close_fds=True,
                    creationflags=subprocess.CREATE_NEW_PROCESS_GROUP
                    | subprocess.CREATE_BREAKAWAY_FROM_JOB,
                )
            else:
                command_name = "start"
                # first parameter is for title, NOT filepath
                command_args = ["", normpath]
                subprocess.Popen(
                    [command_name] + command_args,
                    shell=True,
                    close_fds=True,
                    creationflags=subprocess.CREATE_NEW_PROCESS_GROUP
                    | subprocess.CREATE_BREAKAWAY_FROM_JOB,
                )
        else:
            if sys.platform == "darwin":
                command_name = "open"
                command_args = [_path]
                if file_manager:
                    # will reveal in Finder
                    command_args.append("-R")
            else:
                if file_manager:
                    command_name = "dbus-send"
                    # might not be guaranteed to launch default?
                    command_args = [
                        "--session",
                        "--dest=org.freedesktop.FileManager1",
                        "--type=method_call",
                        "/org/freedesktop/FileManager1",
                        "org.freedesktop.FileManager1.ShowItems",
                        f"array:string:file://{_path}",
                        "string:",
                    ]
                else:
                    command_name = "xdg-open"
                    command_args = [_path]
            command = shutil.which(command_name)
            if command is not None:
                subprocess.Popen([command] + command_args, close_fds=True)
            else:
                logging.info(f"Could not find {command_name} on system PATH")
    except:
        traceback.print_exc()


class FileOpenerHelper:
    def __init__(self, filepath: str | Path):
        """Initialize the FileOpenerHelper.

        Args:
                filepath (str): The path to the file to open.
        """
        self.filepath = str(filepath)

    def set_filepath(self, filepath: str | Path):
        """Set the filepath to open.

        Args:
                filepath (str): The path to the file to open.
        """
        self.filepath = str(filepath)

    def open_file(self):
        """Open the file in the default application."""
        open_file(self.filepath)

    def open_explorer(self):
        """Open the file in the default file explorer."""
        open_file(self.filepath, file_manager=True)


class FileOpenerLabel(QLabel):
    def __init__(self, text, parent=None):
        """Initialize the FileOpenerLabel.

        Args:
                text (str): The text to display.
                parent (QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(text, parent)

    def setFilePath(self, filepath):
        """Set the filepath to open.

        Args:
                filepath (str): The path to the file to open.
        """
        self.filepath = filepath

    def mousePressEvent(self, event):
        """Handle mouse press events.

        On a left click, open the file in the default file explorer. On a right click, show a context menu.

        Args:
                event (QMouseEvent): The mouse press event.
        """
        super().mousePressEvent(event)

        if event.button() == Qt.LeftButton:
            opener = FileOpenerHelper(self.filepath)
            opener.open_explorer()
        elif event.button() == Qt.RightButton:
            # Show context menu
            pass



from typing import TypedDict
from typing_extensions import NotRequired


class JsonLibary(TypedDict("", {"ts-version": str})):
    # "ts-version": str
    tags: "list[JsonTag]"
    collations: "list[JsonCollation]"
    fields: list  # TODO
    macros: "list[JsonMacro]"
    entries: "list[JsonEntry]"
    ext_list: list[str]
    is_exclude_list: bool
    ignored_extensions: NotRequired[list[str]]  # deprecated


class JsonBase(TypedDict):
    id: int


class JsonTag(JsonBase, total=False):
    name: str
    aliases: list[str]
    color: str
    shorthand: str
    subtag_ids: list[int]


class JsonCollation(JsonBase, total=False):
    title: str
    e_ids_and_pages: list[list[int]]
    sort_order: str
    cover_id: int


class JsonEntry(JsonBase, total=False):
    filename: str
    path: str
    fields: list[dict]  # TODO


class JsonMacro(JsonBase, total=False): ...  # TODO



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio

import logging
from pathlib import Path
import time
import typing
from datetime import datetime as dt

import cv2
import rawpy
from PIL import Image, UnidentifiedImageError
from PIL.Image import DecompressionBombError
from PySide6.QtCore import Signal, Qt, QSize
from PySide6.QtGui import QResizeEvent, QAction
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QScrollArea,
    QFrame,
    QSplitter,
    QSizePolicy,
    QMessageBox,
)
from humanfriendly import format_size

from src.core.enums import SettingItems, Theme
from src.core.library import Entry, ItemType, Library
from src.core.constants import VIDEO_TYPES, IMAGE_TYPES, RAW_IMAGE_TYPES, TS_FOLDER_NAME
from src.qt.helpers.file_opener import FileOpenerLabel, FileOpenerHelper, open_file
from src.qt.modals.add_field import AddFieldModal
from src.qt.widgets.thumb_renderer import ThumbRenderer
from src.qt.widgets.fields import FieldContainer
from src.qt.widgets.tag_box import TagBoxWidget
from src.qt.widgets.text import TextWidget
from src.qt.widgets.panel import PanelModal
from src.qt.widgets.text_box_edit import EditTextBox
from src.qt.widgets.text_line_edit import EditTextLine
from src.qt.helpers.qbutton_wrapper import QPushButtonWrapper
from src.qt.widgets.video_player import VideoPlayer


# Only import for type checking/autocompletion, will not be imported at runtime.
if typing.TYPE_CHECKING:
    from src.qt.ts_qt import QtDriver

ERROR = "[ERROR]"
WARNING = "[WARNING]"
INFO = "[INFO]"

logging.basicConfig(format="%(message)s", level=logging.INFO)


class PreviewPanel(QWidget):
    """The Preview Panel Widget."""

    tags_updated = Signal()

    def __init__(self, library: Library, driver: "QtDriver"):
        super().__init__()
        self.is_connected = False
        self.lib = library
        self.driver: QtDriver = driver
        self.initialized = False
        self.isOpen: bool = False
        # self.filepath = None
        # self.item = None # DEPRECATED, USE self.selected
        self.common_fields: list = []
        self.mixed_fields: list = []
        self.selected: list[tuple[ItemType, int]] = []  # New way of tracking items
        self.tag_callback = None
        self.containers: list[QWidget] = []

        self.img_button_size: tuple[int, int] = (266, 266)
        self.image_ratio: float = 1.0

        self.image_container = QWidget()
        image_layout = QHBoxLayout(self.image_container)
        image_layout.setContentsMargins(0, 0, 0, 0)

        self.open_file_action = QAction("Open file", self)
        self.open_explorer_action = QAction("Open file in explorer", self)

        self.preview_img = QPushButtonWrapper()
        self.preview_img.setMinimumSize(*self.img_button_size)
        self.preview_img.setFlat(True)
        self.preview_img.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)

        self.preview_img.addAction(self.open_file_action)
        self.preview_img.addAction(self.open_explorer_action)
        self.preview_vid = VideoPlayer(driver)
        self.preview_vid.hide()
        self.thumb_renderer = ThumbRenderer()
        self.thumb_renderer.updated.connect(
            lambda ts, i, s: (self.preview_img.setIcon(i))
        )
        self.thumb_renderer.updated_ratio.connect(
            lambda ratio: (
                self.set_image_ratio(ratio),
                self.update_image_size(
                    (
                        self.image_container.size().width(),
                        self.image_container.size().height(),
                    ),
                    ratio,
                ),
            )
        )

        image_layout.addWidget(self.preview_img)
        image_layout.setAlignment(self.preview_img, Qt.AlignmentFlag.AlignCenter)
        image_layout.addWidget(self.preview_vid)
        image_layout.setAlignment(self.preview_vid, Qt.AlignmentFlag.AlignCenter)
        self.image_container.setMinimumSize(*self.img_button_size)
        self.file_label = FileOpenerLabel("Filename")
        self.file_label.setWordWrap(True)
        self.file_label.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse
        )
        self.file_label.setStyleSheet("font-weight: bold; font-size: 12px")

        self.dimensions_label = QLabel("Dimensions")
        self.dimensions_label.setWordWrap(True)
        # self.dim_label.setTextInteractionFlags(
        # 	Qt.TextInteractionFlag.TextSelectableByMouse)

        properties_style = (
            f"background-color:{Theme.COLOR_BG.value};"
            f"font-family:Oxanium;"
            f"font-weight:bold;"
            f"font-size:12px;"
            f"border-radius:6px;"
            f"padding-top: 4px;"
            f"padding-right: 1px;"
            f"padding-bottom: 1px;"
            f"padding-left: 1px;"
        )

        self.dimensions_label.setStyleSheet(properties_style)

        self.scroll_layout = QVBoxLayout()
        self.scroll_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.scroll_layout.setContentsMargins(6, 1, 6, 6)

        scroll_container: QWidget = QWidget()
        scroll_container.setObjectName("entryScrollContainer")
        scroll_container.setLayout(self.scroll_layout)

        info_section = QWidget()
        info_layout = QVBoxLayout(info_section)
        info_layout.setContentsMargins(0, 0, 0, 0)
        info_layout.setSpacing(6)

        scroll_area = QScrollArea()
        scroll_area.setObjectName("entryScrollArea")
        scroll_area.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        scroll_area.setWidgetResizable(True)
        scroll_area.setFrameShadow(QFrame.Shadow.Plain)
        scroll_area.setFrameShape(QFrame.Shape.NoFrame)
        # NOTE: I would rather have this style applied to the scroll_area
        # background and NOT the scroll container background, so that the
        # rounded corners are maintained when scrolling. I was unable to
        # find the right trick to only select that particular element.
        scroll_area.setStyleSheet(
            "QWidget#entryScrollContainer{"
            f"background: {Theme.COLOR_BG.value};"
            "border-radius:6px;"
            "}"
        )
        scroll_area.setWidget(scroll_container)

        info_layout.addWidget(self.file_label)
        info_layout.addWidget(self.dimensions_label)
        info_layout.addWidget(scroll_area)

        # keep list of rendered libraries to avoid needless re-rendering
        self.render_libs: set = set()
        self.libs_layout = QVBoxLayout()
        self.fill_libs_widget(self.libs_layout)

        self.libs_flow_container: QWidget = QWidget()
        self.libs_flow_container.setObjectName("librariesList")
        self.libs_flow_container.setLayout(self.libs_layout)
        self.libs_flow_container.setSizePolicy(
            QSizePolicy.Preferred,  # type: ignore
            QSizePolicy.Maximum,  # type: ignore
        )

        # set initial visibility based on settings
        if not self.driver.settings.value(
            SettingItems.WINDOW_SHOW_LIBS, True, type=bool
        ):
            self.libs_flow_container.hide()

        splitter = QSplitter()
        splitter.setOrientation(Qt.Orientation.Vertical)
        splitter.setHandleWidth(12)
        splitter.splitterMoved.connect(
            lambda: self.update_image_size(
                (
                    self.image_container.size().width(),
                    self.image_container.size().height(),
                )
            )
        )

        splitter.addWidget(self.image_container)
        splitter.addWidget(info_section)
        splitter.addWidget(self.libs_flow_container)
        splitter.setStretchFactor(1, 2)

        self.afb_container = QWidget()
        self.afb_layout = QVBoxLayout(self.afb_container)
        self.afb_layout.setContentsMargins(0, 12, 0, 0)

        self.add_field_button = QPushButtonWrapper()
        self.add_field_button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.add_field_button.setMinimumSize(96, 28)
        self.add_field_button.setMaximumSize(96, 28)
        self.add_field_button.setText("Add Field")
        self.afb_layout.addWidget(self.add_field_button)
        self.afm = AddFieldModal(self.lib)
        self.place_add_field_button()
        self.update_image_size(
            (self.image_container.size().width(), self.image_container.size().height())
        )

        root_layout = QHBoxLayout(self)
        root_layout.setContentsMargins(0, 0, 0, 0)
        root_layout.addWidget(splitter)

    def fill_libs_widget(self, layout: QVBoxLayout):
        settings = self.driver.settings
        settings.beginGroup(SettingItems.LIBS_LIST)
        lib_items: dict[str, tuple[str, str]] = {}
        for item_tstamp in settings.allKeys():
            val: str = settings.value(item_tstamp)  # type: ignore
            cut_val = val
            if len(val) > 45:
                cut_val = f"{val[0:10]} ... {val[-10:]}"
            lib_items[item_tstamp] = (val, cut_val)

        settings.endGroup()

        new_keys = set(lib_items.keys())
        if new_keys == self.render_libs:
            # no need to re-render
            return

        # sort lib_items by the key
        libs_sorted = sorted(lib_items.items(), key=lambda item: item[0], reverse=True)

        self.render_libs = new_keys
        self._fill_libs_widget(libs_sorted, layout)

    def _fill_libs_widget(
        self, libraries: list[tuple[str, tuple[str, str]]], layout: QVBoxLayout
    ):
        def clear_layout(layout_item: QVBoxLayout):
            for i in reversed(range(layout_item.count())):
                child = layout_item.itemAt(i)
                if child.widget() is not None:
                    child.widget().deleteLater()
                elif child.layout() is not None:
                    clear_layout(child.layout())  # type: ignore

        # remove any potential previous items
        clear_layout(layout)

        label = QLabel("Recent Libraries")
        label.setAlignment(Qt.AlignCenter)  # type: ignore

        row_layout = QHBoxLayout()
        row_layout.addWidget(label)
        layout.addLayout(row_layout)

        def set_button_style(
            btn: QPushButtonWrapper | QPushButton, extras: list[str] | None = None
        ):
            base_style = [
                f"background-color:{Theme.COLOR_BG.value};",
                "border-radius:6px;",
                "text-align: left;",
                "padding-top: 3px;",
                "padding-left: 6px;",
                "padding-bottom: 4px;",
            ]

            full_style_rows = base_style + (extras or [])

            btn.setStyleSheet(
                (
                    "QPushButton{"
                    f"{''.join(full_style_rows)}"
                    "}"
                    f"QPushButton::hover{{background-color:{Theme.COLOR_HOVER.value};}}"
                    f"QPushButton::pressed{{background-color:{Theme.COLOR_PRESSED.value};}}"
                    f"QPushButton::disabled{{background-color:{Theme.COLOR_DISABLED_BG.value};}}"
                )
            )
            btn.setCursor(Qt.CursorShape.PointingHandCursor)

        for item_key, (full_val, cut_val) in libraries:
            button = QPushButton(text=cut_val)
            button.setObjectName(f"path{item_key}")

            lib = Path(full_val)
            if not lib.exists() or not (lib / TS_FOLDER_NAME).exists():
                button.setDisabled(True)
                button.setToolTip("Location is missing")

            def open_library_button_clicked(path):
                return lambda: self.driver.open_library(Path(path))

            button.clicked.connect(open_library_button_clicked(full_val))
            set_button_style(button)
            button_remove = QPushButton("➖")
            button_remove.setCursor(Qt.CursorShape.PointingHandCursor)
            button_remove.setFixedWidth(30)
            set_button_style(button_remove)

            def remove_recent_library_clicked(key: str):
                return lambda: (
                    self.driver.remove_recent_library(key),
                    self.fill_libs_widget(self.libs_layout),
                )

            button_remove.clicked.connect(remove_recent_library_clicked(item_key))

            row_layout = QHBoxLayout()
            row_layout.addWidget(button)
            row_layout.addWidget(button_remove)

            layout.addLayout(row_layout)

    def resizeEvent(self, event: QResizeEvent) -> None:
        self.update_image_size(
            (self.image_container.size().width(), self.image_container.size().height())
        )
        return super().resizeEvent(event)

    def get_preview_size(self) -> tuple[int, int]:
        return (
            self.image_container.size().width(),
            self.image_container.size().height(),
        )

    def set_image_ratio(self, ratio: float):
        # logging.info(f'Updating Ratio to: {ratio} #####################################################')
        self.image_ratio = ratio

    def update_image_size(self, size: tuple[int, int], ratio: float = None):
        if ratio:
            self.set_image_ratio(ratio)
        # self.img_button_size = size
        # logging.info(f'')
        # self.preview_img.setMinimumSize(64,64)

        adj_width: float = size[0]
        adj_height: float = size[1]
        # Landscape
        if self.image_ratio > 1:
            # logging.info('Landscape')
            adj_height = size[0] * (1 / self.image_ratio)
        # Portrait
        elif self.image_ratio <= 1:
            # logging.info('Portrait')
            adj_width = size[1] * self.image_ratio

        if adj_width > size[0]:
            adj_height = adj_height * (size[0] / adj_width)
            adj_width = size[0]
        elif adj_height > size[1]:
            adj_width = adj_width * (size[1] / adj_height)
            adj_height = size[1]

        # adj_width = min(adj_width, self.image_container.size().width())
        # adj_height = min(adj_width, self.image_container.size().height())

        # self.preview_img.setMinimumSize(s)
        # self.preview_img.setMaximumSize(s_max)
        adj_size = QSize(int(adj_width), int(adj_height))
        self.img_button_size = (int(adj_width), int(adj_height))
        self.preview_img.setMaximumSize(adj_size)
        self.preview_img.setIconSize(adj_size)
        self.preview_vid.resizeVideo(adj_size)
        self.preview_vid.setMaximumSize(adj_size)
        self.preview_vid.setMinimumSize(adj_size)
        # self.preview_img.setMinimumSize(adj_size)

        # if self.preview_img.iconSize().toTuple()[0] < self.preview_img.size().toTuple()[0] + 10:
        # 	if type(self.item) == Entry:
        # 		filepath = os.path.normpath(f'{self.lib.library_dir}/{self.item.path}/{self.item.filename}')
        # 		self.thumb_renderer.render(time.time(), filepath, self.preview_img.size().toTuple(), self.devicePixelRatio(),update_on_ratio_change=True)

        # logging.info(f' Img Aspect Ratio: {self.image_ratio}')
        # logging.info(f'  Max Button Size: {size}')
        # logging.info(f'Container Size: {(self.image_container.size().width(), self.image_container.size().height())}')
        # logging.info(f'Final Button Size: {(adj_width, adj_height)}')
        # logging.info(f'')
        # logging.info(f'  Icon Size: {self.preview_img.icon().actualSize().toTuple()}')
        # logging.info(f'Button Size: {self.preview_img.size().toTuple()}')

    def place_add_field_button(self):
        self.scroll_layout.addWidget(self.afb_container)
        self.scroll_layout.setAlignment(
            self.afb_container, Qt.AlignmentFlag.AlignHCenter
        )

        if self.afm.is_connected:
            self.afm.done.disconnect()
        if self.add_field_button.is_connected:
            self.add_field_button.clicked.disconnect()

        # self.afm.done.connect(lambda f: (self.lib.add_field_to_entry(self.selected[0][1], f), self.update_widgets()))
        self.afm.done.connect(
            lambda f: (self.add_field_to_selected(f), self.update_widgets())
        )
        self.afm.is_connected = True
        self.add_field_button.clicked.connect(self.afm.show)

    def add_field_to_selected(self, field_id: int):
        """Adds an entry field to one or more selected items."""
        added = set()
        for item_pair in self.selected:
            if item_pair[0] == ItemType.ENTRY and item_pair[1] not in added:
                self.lib.add_field_to_entry(item_pair[1], field_id)
                added.add(item_pair[1])

    # def update_widgets(self, item: Union[Entry, Collation, Tag]):
    def update_widgets(self):
        """
        Renders the panel's widgets with the newest data from the Library.
        """
        logging.info(f"[ENTRY PANEL] UPDATE WIDGETS ({self.driver.selected})")
        self.isOpen = True
        # self.tag_callback = tag_callback if tag_callback else None
        window_title = ""

        # update list of libraries
        self.fill_libs_widget(self.libs_layout)

        # 0 Selected Items
        if not self.driver.selected:
            if self.selected or not self.initialized:
                self.file_label.setText("No Items Selected")
                self.file_label.setFilePath("")
                self.file_label.setCursor(Qt.CursorShape.ArrowCursor)

                self.dimensions_label.setText("")
                self.preview_img.setContextMenuPolicy(
                    Qt.ContextMenuPolicy.NoContextMenu
                )
                self.preview_img.setCursor(Qt.CursorShape.ArrowCursor)

                ratio: float = self.devicePixelRatio()
                self.thumb_renderer.render(
                    time.time(),
                    "",
                    (512, 512),
                    ratio,
                    True,
                    update_on_ratio_change=True,
                )
                if self.preview_img.is_connected:
                    self.preview_img.clicked.disconnect()
                for i, c in enumerate(self.containers):
                    c.setHidden(True)
            self.preview_img.show()
            self.preview_vid.stop()
            self.preview_vid.hide()
            self.selected = list(self.driver.selected)
            self.add_field_button.setHidden(True)

        # 1 Selected Item
        elif len(self.driver.selected) == 1:
            # 1 Selected Entry
            if self.driver.selected[0][0] == ItemType.ENTRY:
                self.preview_img.show()
                self.preview_vid.stop()
                self.preview_vid.hide()
                item: Entry = self.lib.get_entry(self.driver.selected[0][1])
                # If a new selection is made, update the thumbnail and filepath.
                if not self.selected or self.selected != self.driver.selected:
                    filepath = self.lib.library_dir / item.path / item.filename
                    self.file_label.setFilePath(filepath)
                    window_title = str(filepath)
                    ratio: float = self.devicePixelRatio()
                    self.thumb_renderer.render(
                        time.time(),
                        filepath,
                        (512, 512),
                        ratio,
                        update_on_ratio_change=True,
                    )
                    self.file_label.setText("\u200b".join(str(filepath)))
                    self.file_label.setCursor(Qt.CursorShape.PointingHandCursor)

                    self.preview_img.setContextMenuPolicy(
                        Qt.ContextMenuPolicy.ActionsContextMenu
                    )
                    self.preview_img.setCursor(Qt.CursorShape.PointingHandCursor)

                    self.opener = FileOpenerHelper(filepath)
                    self.open_file_action.triggered.connect(self.opener.open_file)
                    self.open_explorer_action.triggered.connect(
                        self.opener.open_explorer
                    )

                    # TODO: Do this somewhere else, this is just here temporarily.
                    try:
                        image = None
                        if filepath.suffix.lower() in IMAGE_TYPES:
                            image = Image.open(str(filepath))
                        elif filepath.suffix.lower() in RAW_IMAGE_TYPES:
                            try:
                                with rawpy.imread(str(filepath)) as raw:
                                    rgb = raw.postprocess()
                                    image = Image.new(
                                        "L", (rgb.shape[1], rgb.shape[0]), color="black"
                                    )
                            except (
                                rawpy._rawpy.LibRawIOError,
                                rawpy._rawpy.LibRawFileUnsupportedError,
                            ):
                                pass
                        elif filepath.suffix.lower() in VIDEO_TYPES:
                            video = cv2.VideoCapture(str(filepath))
                            if video.get(cv2.CAP_PROP_FRAME_COUNT) <= 0:
                                raise cv2.error("File is invalid or has 0 frames")
                            video.set(cv2.CAP_PROP_POS_FRAMES, 0)
                            success, frame = video.read()
                            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                            image = Image.fromarray(frame)
                            if success:
                                self.preview_img.hide()
                                self.preview_vid.play(
                                    filepath, QSize(image.width, image.height)
                                )
                                self.resizeEvent(
                                    QResizeEvent(
                                        QSize(image.width, image.height),
                                        QSize(image.width, image.height),
                                    )
                                )
                                self.preview_vid.show()

                        # Stats for specific file types are displayed here.
                        if image and filepath.suffix.lower() in (
                            IMAGE_TYPES + VIDEO_TYPES + RAW_IMAGE_TYPES
                        ):
                            self.dimensions_label.setText(
                                f"{filepath.suffix.upper()[1:]}  •  {format_size(filepath.stat().st_size)}\n{image.width} x {image.height} px"
                            )
                        else:
                            self.dimensions_label.setText(
                                f"{filepath.suffix.upper()[1:]}  •  {format_size(filepath.stat().st_size)}"
                            )

                        if not filepath.is_file():
                            raise FileNotFoundError

                    except FileNotFoundError as e:
                        self.dimensions_label.setText(f"{filepath.suffix.upper()[1:]}")
                        logging.info(
                            f"[PreviewPanel][ERROR] Couldn't Render thumbnail for {filepath} (because of {e})"
                        )

                    except (FileNotFoundError, cv2.error) as e:
                        self.dimensions_label.setText(f"{filepath.suffix.upper()}")
                        logging.info(
                            f"[PreviewPanel][ERROR] Couldn't Render thumbnail for {filepath} (because of {e})"
                        )
                    except (
                        UnidentifiedImageError,
                        DecompressionBombError,
                    ) as e:
                        self.dimensions_label.setText(
                            f"{filepath.suffix.upper()[1:]}  •  {format_size(filepath.stat().st_size)}"
                        )
                        logging.info(
                            f"[PreviewPanel][ERROR] Couldn't Render thumbnail for {filepath} (because of {e})"
                        )

                    if self.preview_img.is_connected:
                        self.preview_img.clicked.disconnect()
                    self.preview_img.clicked.connect(
                        lambda checked=False, filepath=filepath: open_file(filepath)
                    )
                    self.preview_img.is_connected = True
                self.selected = list(self.driver.selected)
                for i, f in enumerate(item.fields):
                    self.write_container(i, f)

                # Hide leftover containers
                if len(self.containers) > len(item.fields):
                    for i, c in enumerate(self.containers):
                        if i > (len(item.fields) - 1):
                            c.setHidden(True)

                self.add_field_button.setHidden(False)

            # 1 Selected Collation
            elif self.driver.selected[0][0] == ItemType.COLLATION:
                pass

            # 1 Selected Tag
            elif self.driver.selected[0][0] == ItemType.TAG_GROUP:
                pass

        # Multiple Selected Items
        elif len(self.driver.selected) > 1:
            self.preview_img.show()
            self.preview_vid.stop()
            self.preview_vid.hide()
            if self.selected != self.driver.selected:
                self.file_label.setText(f"{len(self.driver.selected)} Items Selected")
                self.file_label.setCursor(Qt.CursorShape.ArrowCursor)
                self.file_label.setFilePath("")
                self.dimensions_label.setText("")

                self.preview_img.setContextMenuPolicy(
                    Qt.ContextMenuPolicy.NoContextMenu
                )
                self.preview_img.setCursor(Qt.CursorShape.ArrowCursor)

                ratio: float = self.devicePixelRatio()
                self.thumb_renderer.render(
                    time.time(),
                    "",
                    (512, 512),
                    ratio,
                    True,
                    update_on_ratio_change=True,
                )
                if self.preview_img.is_connected:
                    self.preview_img.clicked.disconnect()
                    self.preview_img.is_connected = False

            self.common_fields = []
            self.mixed_fields = []
            for i, item_pair in enumerate(self.driver.selected):
                if item_pair[0] == ItemType.ENTRY:
                    item = self.lib.get_entry(item_pair[1])
                    if i == 0:
                        for f in item.fields:
                            self.common_fields.append(f)
                    else:
                        common_to_remove = []
                        for f in self.common_fields:
                            # Common field found (Same ID, identical content)
                            if f not in item.fields:
                                common_to_remove.append(f)

                                # Mixed field found (Same ID, different content)
                                if self.lib.get_field_index_in_entry(
                                    item, self.lib.get_field_attr(f, "id")
                                ):
                                    # if self.lib.get_field_attr(f, 'type') == ('tag_box'):
                                    # 	pass
                                    # logging.info(f)
                                    # logging.info(type(f))
                                    f_stripped = {
                                        self.lib.get_field_attr(f, "id"): None
                                    }
                                    if f_stripped not in self.mixed_fields and (
                                        f not in self.common_fields
                                        or f in common_to_remove
                                    ):
                                        #  and (f not in self.common_fields or f in common_to_remove)
                                        self.mixed_fields.append(f_stripped)
                        self.common_fields = [
                            f for f in self.common_fields if f not in common_to_remove
                        ]
            order: list[int] = (
                [0]
                + [1, 2]
                + [9, 17, 18, 19, 20]
                + [8, 7, 6]
                + [4]
                + [3, 21]
                + [10, 14, 11, 12, 13, 22]
                + [5]
            )
            self.mixed_fields = sorted(
                self.mixed_fields,
                key=lambda x: order.index(self.lib.get_field_attr(x, "id")),
            )

            self.selected = list(self.driver.selected)
            for i, f in enumerate(self.common_fields):
                logging.info(f"ci:{i}, f:{f}")
                self.write_container(i, f)
            for i, f in enumerate(self.mixed_fields, start=len(self.common_fields)):
                logging.info(f"mi:{i}, f:{f}")
                self.write_container(i, f, mixed=True)

            # Hide leftover containers
            if len(self.containers) > len(self.common_fields) + len(self.mixed_fields):
                for i, c in enumerate(self.containers):
                    if i > (len(self.common_fields) + len(self.mixed_fields) - 1):
                        c.setHidden(True)

            self.add_field_button.setHidden(False)

        self.initialized = True

        # # Uninitialized or New Item:
        # if not self.item or self.item.id != item.id:
        # 	# logging.info(f'Uninitialized or New Item ({item.id})')
        # 	if type(item) == Entry:
        # 		# New Entry: Render preview and update filename label
        # 		filepath = os.path.normpath(f'{self.lib.library_dir}/{item.path}/{item.filename}')
        # 		window_title = filepath
        # 		ratio: float = self.devicePixelRatio()
        # 		self.thumb_renderer.render(time.time(), filepath, (512, 512), ratio,update_on_ratio_change=True)
        # 		self.file_label.setText("\u200b".join(filepath))

        # 		# TODO: Deal with this later.
        # 		# https://stackoverflow.com/questions/64252654/pyqt5-drag-and-drop-into-system-file-explorer-with-delayed-encoding
        # 		# https://doc.qt.io/qtforpython-5/PySide2/QtCore/QMimeData.html#more
        # 		# drag = QDrag(self.preview_img)
        # 		# mime = QMimeData()
        # 		# mime.setUrls([filepath])
        # 		# drag.setMimeData(mime)
        # 		# drag.exec_(Qt.DropAction.CopyAction)

        # 		try:
        # 			self.preview_img.clicked.disconnect()
        # 		except RuntimeError:
        # 			pass
        # 		self.preview_img.clicked.connect(
        # 			lambda checked=False, filepath=filepath: open_file(filepath))

        # 		for i, f in enumerate(item.fields):
        # 			self.write_container(item, i, f)

        # 		self.item = item

        # 		# try:
        # 		# 	self.tags_updated.disconnect()
        # 		# except RuntimeError:
        # 		# 	pass
        # 		# if self.tag_callback:
        # 		# 	# logging.info(f'[UPDATE CONTAINER] Updating Callback for {item.id}: {self.tag_callback}')
        # 		# 	self.tags_updated.connect(self.tag_callback)

        # # Initialized, Updating:
        # elif self.item and self.item.id == item.id:
        # 	# logging.info(f'Initialized Item, Updating! ({item.id})')
        # 	for i, f in enumerate(item.fields):
        # 		self.write_container(item, i, f)

        # # Hide leftover containers
        # if len(self.containers) > len(self.item.fields):
        # 	for i, c in enumerate(self.containers):
        # 		if i > (len(self.item.fields) - 1):
        # 			c.setHidden(True)

        self.setWindowTitle(window_title)
        self.show()

    def set_tags_updated_slot(self, slot: object):
        """
        Replacement for tag_callback.
        """
        if self.is_connected:
            self.tags_updated.disconnect()

        logging.info("[UPDATE CONTAINER] Setting tags updated slot")
        self.tags_updated.connect(slot)
        self.is_connected = True

    # def write_container(self, item:Union[Entry, Collation, Tag], index, field):
    def write_container(self, index, field, mixed=False):
        """Updates/Creates data for a FieldContainer."""
        # logging.info(f'[ENTRY PANEL] WRITE CONTAINER')
        # Remove 'Add Field' button from scroll_layout, to be re-added later.
        self.scroll_layout.takeAt(self.scroll_layout.count() - 1).widget()
        container: FieldContainer = None
        if len(self.containers) < (index + 1):
            container = FieldContainer()
            self.containers.append(container)
            self.scroll_layout.addWidget(container)
        else:
            container = self.containers[index]
            # container.inner_layout.removeItem(container.inner_layout.itemAt(1))
            # container.setHidden(False)
        if self.lib.get_field_attr(field, "type") == "tag_box":
            # logging.info(f'WRITING TAGBOX FOR ITEM {item.id}')
            container.set_title(self.lib.get_field_attr(field, "name"))
            # container.set_editable(False)
            container.set_inline(False)
            title = f"{self.lib.get_field_attr(field, 'name')} (Tag Box)"
            if not mixed:
                item = self.lib.get_entry(
                    self.selected[0][1]
                )  # TODO TODO TODO: TEMPORARY
                if type(container.get_inner_widget()) == TagBoxWidget:
                    inner_container: TagBoxWidget = container.get_inner_widget()
                    inner_container.set_item(item)
                    inner_container.set_tags(self.lib.get_field_attr(field, "content"))
                    try:
                        inner_container.updated.disconnect()
                    except RuntimeError:
                        pass
                    # inner_container.updated.connect(lambda f=self.filepath, i=item: self.write_container(item, index, field))
                else:
                    inner_container = TagBoxWidget(
                        item,
                        title,
                        index,
                        self.lib,
                        self.lib.get_field_attr(field, "content"),
                        self.driver,
                    )

                    container.set_inner_widget(inner_container)
                inner_container.field = field
                inner_container.updated.connect(
                    lambda: (
                        self.write_container(index, field),
                        self.tags_updated.emit(),
                    )
                )
                # if type(item) == Entry:
                # NOTE: Tag Boxes have no Edit Button (But will when you can convert field types)
                # f'Are you sure you want to remove this \"{self.lib.get_field_attr(field, "name")}\" field?'
                # container.set_remove_callback(lambda: (self.lib.get_entry(item.id).fields.pop(index), self.update_widgets(item)))
                prompt = f'Are you sure you want to remove this "{self.lib.get_field_attr(field, "name")}" field?'
                callback = lambda: (self.remove_field(field), self.update_widgets())
                container.set_remove_callback(
                    lambda: self.remove_message_box(prompt=prompt, callback=callback)
                )
                container.set_copy_callback(None)
                container.set_edit_callback(None)
            else:
                text = "<i>Mixed Data</i>"
                title = f"{self.lib.get_field_attr(field, 'name')} (Wacky Tag Box)"
                inner_container = TextWidget(title, text)
                container.set_inner_widget(inner_container)
                container.set_copy_callback(None)
                container.set_edit_callback(None)
                container.set_remove_callback(None)

            self.tags_updated.emit()
            # self.dynamic_widgets.append(inner_container)
        elif self.lib.get_field_attr(field, "type") in "text_line":
            # logging.info(f'WRITING TEXTLINE FOR ITEM {item.id}')
            container.set_title(self.lib.get_field_attr(field, "name"))
            # container.set_editable(True)
            container.set_inline(False)
            # Normalize line endings in any text content.
            if not mixed:
                text = self.lib.get_field_attr(field, "content").replace("\r", "\n")
            else:
                text = "<i>Mixed Data</i>"
            title = f"{self.lib.get_field_attr(field, 'name')} (Text Line)"
            inner_container = TextWidget(title, text)
            container.set_inner_widget(inner_container)
            # if type(item) == Entry:
            if not mixed:
                modal = PanelModal(
                    EditTextLine(self.lib.get_field_attr(field, "content")),
                    title=title,
                    window_title=f'Edit {self.lib.get_field_attr(field, "name")}',
                    save_callback=(
                        lambda content: (
                            self.update_field(field, content),
                            self.update_widgets(),
                        )
                    ),
                )
                container.set_edit_callback(modal.show)
                prompt = f'Are you sure you want to remove this "{self.lib.get_field_attr(field, "name")}" field?'
                callback = lambda: (self.remove_field(field), self.update_widgets())
                container.set_remove_callback(
                    lambda: self.remove_message_box(prompt=prompt, callback=callback)
                )
                container.set_copy_callback(None)
            else:
                container.set_edit_callback(None)
                container.set_copy_callback(None)
                container.set_remove_callback(None)
            # container.set_remove_callback(lambda: (self.lib.get_entry(item.id).fields.pop(index), self.update_widgets(item)))

        elif self.lib.get_field_attr(field, "type") in "text_box":
            # logging.info(f'WRITING TEXTBOX FOR ITEM {item.id}')
            container.set_title(self.lib.get_field_attr(field, "name"))
            # container.set_editable(True)
            container.set_inline(False)
            # Normalize line endings in any text content.
            if not mixed:
                text = self.lib.get_field_attr(field, "content").replace("\r", "\n")
            else:
                text = "<i>Mixed Data</i>"
            title = f"{self.lib.get_field_attr(field, 'name')} (Text Box)"
            inner_container = TextWidget(title, text)
            container.set_inner_widget(inner_container)
            # if type(item) == Entry:
            if not mixed:
                container.set_copy_callback(None)
                modal = PanelModal(
                    EditTextBox(self.lib.get_field_attr(field, "content")),
                    title=title,
                    window_title=f'Edit {self.lib.get_field_attr(field, "name")}',
                    save_callback=(
                        lambda content: (
                            self.update_field(field, content),
                            self.update_widgets(),
                        )
                    ),
                )
                container.set_edit_callback(modal.show)
                prompt = f'Are you sure you want to remove this "{self.lib.get_field_attr(field, "name")}" field?'
                callback = lambda: (self.remove_field(field), self.update_widgets())
                container.set_remove_callback(
                    lambda: self.remove_message_box(prompt=prompt, callback=callback)
                )
            else:
                container.set_edit_callback(None)
                container.set_copy_callback(None)
                container.set_remove_callback(None)
        elif self.lib.get_field_attr(field, "type") == "collation":
            # logging.info(f'WRITING COLLATION FOR ITEM {item.id}')
            container.set_title(self.lib.get_field_attr(field, "name"))
            # container.set_editable(True)
            container.set_inline(False)
            collation = self.lib.get_collation(
                self.lib.get_field_attr(field, "content")
            )
            title = f"{self.lib.get_field_attr(field, 'name')} (Collation)"
            text = f"{collation.title} ({len(collation.e_ids_and_pages)} Items)"
            if len(self.selected) == 1:
                text += f" - Page {collation.e_ids_and_pages[[x[0] for x in collation.e_ids_and_pages].index(self.selected[0][1])][1]}"
            inner_container = TextWidget(title, text)
            container.set_inner_widget(inner_container)
            # if type(item) == Entry:
            container.set_copy_callback(None)
            # container.set_edit_callback(None)
            # container.set_remove_callback(lambda: (self.lib.get_entry(item.id).fields.pop(index), self.update_widgets(item)))
            prompt = f'Are you sure you want to remove this "{self.lib.get_field_attr(field, "name")}" field?'
            callback = lambda: (self.remove_field(field), self.update_widgets())
            container.set_remove_callback(
                lambda: self.remove_message_box(prompt=prompt, callback=callback)
            )
        elif self.lib.get_field_attr(field, "type") == "datetime":
            # logging.info(f'WRITING DATETIME FOR ITEM {item.id}')
            if not mixed:
                try:
                    container.set_title(self.lib.get_field_attr(field, "name"))
                    # container.set_editable(False)
                    container.set_inline(False)
                    # TODO: Localize this and/or add preferences.
                    date = dt.strptime(
                        self.lib.get_field_attr(field, "content"), "%Y-%m-%d %H:%M:%S"
                    )
                    title = f"{self.lib.get_field_attr(field, 'name')} (Date)"
                    inner_container = TextWidget(title, date.strftime("%D - %r"))
                    container.set_inner_widget(inner_container)
                except:
                    container.set_title(self.lib.get_field_attr(field, "name"))
                    # container.set_editable(False)
                    container.set_inline(False)
                    title = f"{self.lib.get_field_attr(field, 'name')} (Date) (Unknown Format)"
                    inner_container = TextWidget(
                        title, str(self.lib.get_field_attr(field, "content"))
                    )
                # if type(item) == Entry:
                container.set_copy_callback(None)
                container.set_edit_callback(None)
                # container.set_remove_callback(lambda: (self.lib.get_entry(item.id).fields.pop(index), self.update_widgets(item)))
                prompt = f'Are you sure you want to remove this "{self.lib.get_field_attr(field, "name")}" field?'
                callback = lambda: (self.remove_field(field), self.update_widgets())
                container.set_remove_callback(
                    lambda: self.remove_message_box(prompt=prompt, callback=callback)
                )
            else:
                text = "<i>Mixed Data</i>"
                title = f"{self.lib.get_field_attr(field, 'name')} (Wacky Date)"
                inner_container = TextWidget(title, text)
                container.set_inner_widget(inner_container)
                container.set_copy_callback(None)
                container.set_edit_callback(None)
                container.set_remove_callback(None)
        else:
            # logging.info(f'[ENTRY PANEL] Unknown Type: {self.lib.get_field_attr(field, "type")}')
            container.set_title(self.lib.get_field_attr(field, "name"))
            # container.set_editable(False)
            container.set_inline(False)
            title = f"{self.lib.get_field_attr(field, 'name')} (Unknown Field Type)"
            inner_container = TextWidget(
                title, str(self.lib.get_field_attr(field, "content"))
            )
            container.set_inner_widget(inner_container)
            # if type(item) == Entry:
            container.set_copy_callback(None)
            container.set_edit_callback(None)
            # container.set_remove_callback(lambda: (self.lib.get_entry(item.id).fields.pop(index), self.update_widgets(item)))
            prompt = f'Are you sure you want to remove this "{self.lib.get_field_attr(field, "name")}" field?'
            callback = lambda: (self.remove_field(field), self.update_widgets())
            # callback = lambda: (self.lib.get_entry(item.id).fields.pop(index), self.update_widgets())
            container.set_remove_callback(
                lambda: self.remove_message_box(prompt=prompt, callback=callback)
            )
        container.edit_button.setHidden(True)
        container.setHidden(False)
        self.place_add_field_button()

    def remove_field(self, field: dict):
        """Removes a field from all selected Entries, given a field object."""
        for item_pair in self.selected:
            if item_pair[0] == ItemType.ENTRY:
                entry = self.lib.get_entry(item_pair[1])
                try:
                    index = entry.fields.index(field)
                    updated_badges = False
                    if 8 in entry.fields[index].keys() and (
                        1 in entry.fields[index][8] or 0 in entry.fields[index][8]
                    ):
                        updated_badges = True
                    # TODO: Create a proper Library/Entry method to manage fields.
                    entry.fields.pop(index)
                    if updated_badges:
                        self.driver.update_badges()
                except ValueError:
                    logging.info(
                        f"[PREVIEW PANEL][ERROR?] Tried to remove field from Entry ({entry.id}) that never had it"
                    )
                    pass

    def update_field(self, field: dict, content):
        """Removes a field from all selected Entries, given a field object."""
        field = dict(field)
        for item_pair in self.selected:
            if item_pair[0] == ItemType.ENTRY:
                entry = self.lib.get_entry(item_pair[1])
                try:
                    logging.info(field)
                    index = entry.fields.index(field)
                    self.lib.update_entry_field(entry.id, index, content, "replace")
                except ValueError:
                    logging.info(
                        f"[PREVIEW PANEL][ERROR] Tried to update field from Entry ({entry.id}) that never had it"
                    )
                    pass

    def remove_message_box(self, prompt: str, callback: typing.Callable) -> None:
        remove_mb = QMessageBox()
        remove_mb.setText(prompt)
        remove_mb.setWindowTitle("Remove Field")
        remove_mb.setIcon(QMessageBox.Icon.Warning)
        cancel_button = remove_mb.addButton(
            "&Cancel", QMessageBox.ButtonRole.DestructiveRole
        )
        remove_button = remove_mb.addButton(
            "&Remove", QMessageBox.ButtonRole.RejectRole
        )
        # remove_mb.setStandardButtons(QMessageBox.StandardButton.Cancel)
        remove_mb.setDefaultButton(cancel_button)
        remove_mb.setEscapeButton(cancel_button)
        result = remove_mb.exec_()
        # logging.info(result)
        if result == 3:
            callback()



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


from PySide6.QtCore import Qt
from PySide6.QtWidgets import QHBoxLayout, QLabel
from src.qt.widgets.fields import FieldWidget


class TextWidget(FieldWidget):
    def __init__(self, title, text: str) -> None:
        super().__init__(title)
        # self.item = item
        self.setObjectName("textBox")
        # self.setStyleSheet('background-color:purple;')
        self.base_layout = QHBoxLayout()
        self.base_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.base_layout)
        self.text_label = QLabel()
        # self.text_label.textFormat(Qt.TextFormat.RichText)
        self.text_label.setStyleSheet("font-size: 12px")
        self.text_label.setWordWrap(True)
        self.text_label.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse
        )
        self.base_layout.addWidget(self.text_label)
        self.set_text(text)

    def set_text(self, text: str):
        self.text_label.setText(text)



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


import logging
import typing

from PySide6.QtCore import Qt, QThreadPool
from PySide6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QPushButton

from src.core.library import Library
from src.qt.helpers.function_iterator import FunctionIterator
from src.qt.helpers.custom_runnable import CustomRunnable
from src.qt.modals.delete_unlinked import DeleteUnlinkedEntriesModal
from src.qt.modals.relink_unlinked import RelinkUnlinkedEntries
from src.qt.modals.merge_dupe_entries import MergeDuplicateEntries
from src.qt.widgets.progress import ProgressWidget

# Only import for type checking/autocompletion, will not be imported at runtime.
if typing.TYPE_CHECKING:
    from src.qt.ts_qt import QtDriver


ERROR = "[ERROR]"
WARNING = "[WARNING]"
INFO = "[INFO]"

logging.basicConfig(format="%(message)s", level=logging.INFO)


class FixUnlinkedEntriesModal(QWidget):
    def __init__(self, library: "Library", driver: "QtDriver"):
        super().__init__()
        self.lib = library
        self.driver = driver
        self.missing_count = -1
        self.dupe_count = -1
        self.setWindowTitle("Fix Unlinked Entries")
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setMinimumSize(400, 300)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 6, 6, 6)

        self.unlinked_desc_widget = QLabel()
        self.unlinked_desc_widget.setObjectName("unlinkedDescriptionLabel")
        self.unlinked_desc_widget.setWordWrap(True)
        self.unlinked_desc_widget.setStyleSheet("text-align:left;")
        self.unlinked_desc_widget.setText(
            """Each library entry is linked to a file in one of your directories. If a file linked to an entry is moved or deleted outside of TagStudio, it is then considered unlinked. Unlinked entries may be automatically relinked via searching your directories, manually relinked by the user, or deleted if desired."""
        )

        self.dupe_desc_widget = QLabel()
        self.dupe_desc_widget.setObjectName("dupeDescriptionLabel")
        self.dupe_desc_widget.setWordWrap(True)
        self.dupe_desc_widget.setStyleSheet("text-align:left;")
        self.dupe_desc_widget.setText(
            """Duplicate entries are defined as multiple entries which point to the same file on disk. Merging these will combine the tags and metadata from all duplicates into a single consolidated entry. These are not to be confused with "duplicate files", which are duplicates of your files themselves outside of TagStudio."""
        )

        self.missing_count_label = QLabel()
        self.missing_count_label.setObjectName("missingCountLabel")
        self.missing_count_label.setStyleSheet("font-weight:bold;" "font-size:14px;")
        self.missing_count_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.dupe_count_label = QLabel()
        self.dupe_count_label.setObjectName("dupeCountLabel")
        self.dupe_count_label.setStyleSheet("font-weight:bold;" "font-size:14px;")
        self.dupe_count_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.refresh_unlinked_button = QPushButton()
        self.refresh_unlinked_button.setText("&Refresh All")
        self.refresh_unlinked_button.clicked.connect(
            lambda: self.refresh_missing_files()
        )

        self.merge_class = MergeDuplicateEntries(self.lib, self.driver)
        self.relink_class = RelinkUnlinkedEntries(self.lib, self.driver)

        self.search_button = QPushButton()
        self.search_button.setText("&Search && Relink")
        self.relink_class.done.connect(
            lambda: self.refresh_and_repair_dupe_entries(self.merge_class)
        )
        self.search_button.clicked.connect(lambda: self.relink_class.repair_entries())

        self.refresh_dupe_button = QPushButton()
        self.refresh_dupe_button.setText("Refresh Duplicate Entries")
        self.refresh_dupe_button.clicked.connect(lambda: self.refresh_dupe_entries())

        self.merge_dupe_button = QPushButton()
        self.merge_dupe_button.setText("&Merge Duplicate Entries")
        self.merge_class.done.connect(lambda: self.set_dupe_count(-1))
        self.merge_class.done.connect(lambda: self.set_missing_count(-1))
        self.merge_class.done.connect(lambda: self.driver.filter_items())
        self.merge_dupe_button.clicked.connect(lambda: self.merge_class.merge_entries())

        self.manual_button = QPushButton()
        self.manual_button.setText("&Manual Relink")

        self.delete_button = QPushButton()
        self.delete_modal = DeleteUnlinkedEntriesModal(self.lib, self.driver)
        self.delete_modal.done.connect(
            lambda: self.set_missing_count(len(self.lib.missing_files))
        )
        self.delete_modal.done.connect(lambda: self.driver.update_thumbs())
        self.delete_button.setText("De&lete Unlinked Entries")
        self.delete_button.clicked.connect(lambda: self.delete_modal.show())

        self.button_container = QWidget()
        self.button_layout = QHBoxLayout(self.button_container)
        self.button_layout.setContentsMargins(6, 6, 6, 6)
        self.button_layout.addStretch(1)

        self.done_button = QPushButton()
        self.done_button.setText("&Done")
        self.done_button.setDefault(True)
        self.done_button.clicked.connect(self.hide)
        self.button_layout.addWidget(self.done_button)

        self.root_layout.addWidget(self.missing_count_label)
        self.root_layout.addWidget(self.unlinked_desc_widget)
        self.root_layout.addWidget(self.refresh_unlinked_button)
        self.root_layout.addWidget(self.search_button)
        self.manual_button.setHidden(True)
        self.root_layout.addWidget(self.manual_button)
        self.root_layout.addWidget(self.delete_button)
        self.root_layout.addStretch(1)
        self.root_layout.addWidget(self.dupe_count_label)
        self.root_layout.addWidget(self.dupe_desc_widget)
        self.root_layout.addWidget(self.refresh_dupe_button)
        self.root_layout.addWidget(self.merge_dupe_button)
        self.root_layout.addStretch(2)
        self.root_layout.addWidget(self.button_container)

        self.set_missing_count(self.missing_count)
        self.set_dupe_count(self.dupe_count)

    def refresh_missing_files(self):
        iterator = FunctionIterator(self.lib.refresh_missing_files)
        pw = ProgressWidget(
            window_title="Scanning Library",
            label_text="Scanning Library for Unlinked Entries...",
            cancel_button_text=None,
            minimum=0,
            maximum=len(self.lib.entries),
        )
        pw.show()
        iterator.value.connect(lambda v: pw.update_progress(v + 1))
        r = CustomRunnable(lambda: iterator.run())
        QThreadPool.globalInstance().start(r)
        r.done.connect(
            lambda: (
                pw.hide(),
                pw.deleteLater(),
                self.set_missing_count(len(self.lib.missing_files)),
                self.delete_modal.refresh_list(),
                self.refresh_dupe_entries(),
            )
        )

    def refresh_dupe_entries(self):
        iterator = FunctionIterator(self.lib.refresh_dupe_entries)
        pw = ProgressWidget(
            window_title="Scanning Library",
            label_text="Scanning Library for Duplicate Entries...",
            cancel_button_text=None,
            minimum=0,
            maximum=len(self.lib.entries),
        )
        pw.show()
        iterator.value.connect(lambda v: pw.update_progress(v + 1))
        r = CustomRunnable(lambda: iterator.run())
        QThreadPool.globalInstance().start(r)
        r.done.connect(
            lambda: (
                pw.hide(),
                pw.deleteLater(),
                self.set_dupe_count(len(self.lib.dupe_entries)),
            )
        )

    def refresh_and_repair_dupe_entries(self, merge_class: MergeDuplicateEntries):
        iterator = FunctionIterator(self.lib.refresh_dupe_entries)
        pw = ProgressWidget(
            window_title="Scanning Library",
            label_text="Scanning Library for Duplicate Entries...",
            cancel_button_text=None,
            minimum=0,
            maximum=len(self.lib.entries),
        )
        pw.show()
        iterator.value.connect(lambda v: pw.update_progress(v + 1))
        r = CustomRunnable(lambda: iterator.run())
        QThreadPool.globalInstance().start(r)
        r.done.connect(
            lambda: (
                pw.hide(),  # type: ignore
                pw.deleteLater(),  # type: ignore
                self.set_dupe_count(len(self.lib.dupe_entries)),
                merge_class.merge_entries(),
            )
        )

    def set_missing_count(self, count: int):
        self.missing_count = count
        if self.missing_count < 0:
            self.search_button.setDisabled(True)
            self.delete_button.setDisabled(True)
            self.missing_count_label.setText("Unlinked Entries: N/A")
        elif self.missing_count == 0:
            self.search_button.setDisabled(True)
            self.delete_button.setDisabled(True)
            self.missing_count_label.setText(f"Unlinked Entries: {count}")
        else:
            self.search_button.setDisabled(False)
            self.delete_button.setDisabled(False)
            self.missing_count_label.setText(f"Unlinked Entries: {count}")

    def set_dupe_count(self, count: int):
        self.dupe_count = count
        if self.dupe_count < 0:
            self.dupe_count_label.setText("Duplicate Entries: N/A")
            self.merge_dupe_button.setDisabled(True)
        elif self.dupe_count == 0:
            self.dupe_count_label.setText(f"Duplicate Entries: {count}")
            self.merge_dupe_button.setDisabled(True)
        else:
            self.dupe_count_label.setText(f"Duplicate Entries: {count}")
            self.merge_dupe_button.setDisabled(False)



# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


import logging
import math
import typing

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QScrollArea,
    QFrame,
)

from src.core.enums import FieldID
from src.core.library import Library, Tag
from src.core.palette import ColorType, get_tag_color
from src.qt.flowlayout import FlowLayout

# Only import for type checking/autocompletion, will not be imported at runtime.
if typing.TYPE_CHECKING:
    from src.qt.ts_qt import QtDriver


ERROR = f"[ERROR]"
WARNING = f"[WARNING]"
INFO = f"[INFO]"

logging.basicConfig(format="%(message)s", level=logging.INFO)


def folders_to_tags(library: Library):
    logging.info("Converting folders to Tags")
    tree: dict = dict(dirs={})

    def add_tag_to_tree(items: list[Tag]):
        branch = tree
        for tag in items:
            if tag.name not in branch["dirs"]:
                branch["dirs"][tag.name] = dict(dirs={}, tag=tag)
            branch = branch["dirs"][tag.name]

    def add_folders_to_tree(items: list[str]) -> Tag:
        branch: dict = tree
        for folder in items:
            if folder not in branch["dirs"]:
                new_tag = Tag(
                    -1,
                    folder,
                    "",
                    [],
                    ([branch["tag"].id] if "tag" in branch else []),
                    "",
                )
                library.add_tag_to_library(new_tag)
                branch["dirs"][folder] = dict(dirs={}, tag=new_tag)
            branch = branch["dirs"][folder]
        return branch.get("tag")

    for tag in library.tags:
        reversed_tag = reverse_tag(library, tag, None)
        add_tag_to_tree(reversed_tag)

    for entry in library.entries:
        folders = list(entry.path.parts)
        if len(folders) == 1 and folders[0] == "":
            continue
        tag = add_folders_to_tree(folders)
        if tag:
            if not entry.has_tag(library, tag.id):
                entry.add_tag(library, tag.id, FieldID.TAGS)

    logging.info("Done")


def reverse_tag(library: Library, tag: Tag, list: list[Tag]) -> list[Tag]:
    if list is not None:
        list.append(tag)
    else:
        list = [tag]

    if len(tag.subtag_ids) == 0:
        list.reverse()
        return list
    else:
        for subtag_id in tag.subtag_ids:
            subtag = library.get_tag(subtag_id)
        return reverse_tag(library, subtag, list)


# =========== UI ===========


def generate_preview_data(library: Library):
    tree: dict = dict(dirs={}, files=[])

    def add_tag_to_tree(items: list[Tag]):
        branch: dict = tree
        for tag in items:
            if tag.name not in branch["dirs"]:
                branch["dirs"][tag.name] = dict(dirs={}, tag=tag, files=[])
            branch = branch["dirs"][tag.name]

    def add_folders_to_tree(items: list[str]) -> dict:
        branch: dict = tree
        for folder in items:
            if folder not in branch["dirs"]:
                new_tag = Tag(-1, folder, "", [], [], "green")
                branch["dirs"][folder] = dict(dirs={}, tag=new_tag, files=[])
            branch = branch["dirs"][folder]
        return branch

    for tag in library.tags:
        reversed_tag = reverse_tag(library, tag, None)
        add_tag_to_tree(reversed_tag)

    for entry in library.entries:
        folders = list(entry.path.parts)
        if len(folders) == 1 and folders[0] == "":
            continue
        branch = add_folders_to_tree(folders)
        if branch:
            field_indexes = library.get_field_index_in_entry(entry, 6)
            has_tag = False
            for index in field_indexes:
                content = library.get_field_attr(entry.fields[index], "content")
                for tag_id in content:
                    tag = library.get_tag(tag_id)
                    if tag.name == branch["tag"].name:
                        has_tag = True
                        break
            if not has_tag:
                branch["files"].append(entry.filename)

    def cut_branches_adding_nothing(branch: dict):
        folders = set(branch["dirs"].keys())
        for folder in folders:
            cut = cut_branches_adding_nothing(branch["dirs"][folder])
            if cut:
                branch["dirs"].pop(folder)

        if "tag" not in branch:
            return
        if branch["tag"].id == -1 or len(branch["files"]) > 0:  # Needs to be first
            return False
        if len(branch["dirs"].keys()) == 0:
            return True

    cut_branches_adding_nothing(tree)

    return tree


class FoldersToTagsModal(QWidget):
    # done = Signal(int)
    def __init__(self, library: "Library", driver: "QtDriver"):
        super().__init__()
        self.library = library
        self.driver = driver
        self.count = -1
        self.filename = ""

        self.setWindowTitle(f"Create Tags From Folders")
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setMinimumSize(640, 640)
        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(6, 6, 6, 6)

        self.title_widget = QLabel()
        self.title_widget.setObjectName("title")
        self.title_widget.setWordWrap(True)
        self.title_widget.setStyleSheet(
            "font-weight:bold;" "font-size:14px;" "padding-top: 6px"
        )
        self.title_widget.setText("Create Tags From Folders")
        self.title_widget.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.desc_widget = QLabel()
        self.desc_widget.setObjectName("descriptionLabel")
        self.desc_widget.setWordWrap(True)
        self.desc_widget.setText(
            """Creates tags based on your folder structure and applies them to your entries.\n The structure below shows all the tags that will be created and what entries they will be applied to."""
        )
        self.desc_widget.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.open_close_button_w = QWidget()
        self.open_close_button_layout = QHBoxLayout(self.open_close_button_w)

        self.open_all_button = QPushButton()
        self.open_all_button.setText("Open All")
        self.open_all_button.clicked.connect(lambda: self.set_all_branches(False))
        self.close_all_button = QPushButton()
        self.close_all_button.setText("Close All")
        self.close_all_button.clicked.connect(lambda: self.set_all_branches(True))

        self.open_close_button_layout.addWidget(self.open_all_button)
        self.open_close_button_layout.addWidget(self.close_all_button)

        self.scroll_contents = QWidget()
        self.scroll_layout = QVBoxLayout(self.scroll_contents)
        self.scroll_layout.setContentsMargins(6, 0, 6, 0)
        self.scroll_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        self.scroll_area = QScrollArea()
        self.scroll_area.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOn
        )
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setFrameShadow(QFrame.Shadow.Plain)
        self.scroll_area.setFrameShape(QFrame.Shape.NoFrame)
        self.scroll_area.setWidget(self.scroll_contents)

        self.apply_button = QPushButton()
        self.apply_button.setText("&Apply")
        self.apply_button.setMinimumWidth(100)
        self.apply_button.clicked.connect(self.on_apply)

        self.showEvent = self.on_open  # type: ignore

        self.root_layout.addWidget(self.title_widget)
        self.root_layout.addWidget(self.desc_widget)
        self.root_layout.addWidget(self.open_close_button_w)
        self.root_layout.addWidget(self.scroll_area)
        self.root_layout.addWidget(
            self.apply_button, alignment=Qt.AlignmentFlag.AlignCenter
        )

    def on_apply(self, event):
        folders_to_tags(self.library)
        self.close()
        self.driver.preview_panel.update_widgets()

    def on_open(self, event):
        for i in reversed(range(self.scroll_layout.count())):
            self.scroll_layout.itemAt(i).widget().setParent(None)

        data = generate_preview_data(self.library)

        for folder in data["dirs"].values():
            test = TreeItem(folder, None)
            self.scroll_layout.addWidget(test)

    def set_all_branches(self, hidden: bool):
        for i in reversed(range(self.scroll_layout.count())):
            child = self.scroll_layout.itemAt(i).widget()
            if type(child) == TreeItem:
                child.set_all_branches(hidden)


class TreeItem(QWidget):
    def __init__(self, data: dict, parentTag: Tag):
        super().__init__()

        self.setStyleSheet("QLabel{font-size: 13px}")

        self.root_layout = QVBoxLayout(self)
        self.root_layout.setContentsMargins(20, 0, 0, 0)
        self.root_layout.setSpacing(1)

        self.test = QWidget()
        self.root_layout.addWidget(self.test)

        self.tag_layout = FlowLayout(self.test)

        self.label = QLabel()
        self.tag_layout.addWidget(self.label)
        self.tag_widget = ModifiedTagWidget(data["tag"], parentTag)
        self.tag_widget.bg_button.clicked.connect(lambda: self.hide_show())
        self.tag_layout.addWidget(self.tag_widget)

        self.children_widget = QWidget()
        self.children_layout = QVBoxLayout(self.children_widget)
        self.root_layout.addWidget(self.children_widget)

        self.populate(data)

    def hide_show(self):
        self.children_widget.setHidden(not self.children_widget.isHidden())
        self.label.setText(">" if self.children_widget.isHidden() else "v")

    def populate(self, data: dict):
        for folder in data["dirs"].values():
            item = TreeItem(folder, data["tag"])
            self.children_layout.addWidget(item)
        for file in data["files"]:
            label = QLabel()
            label.setText("    ->  " + str(file))
            self.children_layout.addWidget(label)

        if len(data["files"]) == 0 and len(data["dirs"].values()) == 0:
            self.hide_show()
        else:
            self.label.setText("v")

    def set_all_branches(self, hidden: bool):
        for i in reversed(range(self.children_layout.count())):
            child = self.children_layout.itemAt(i).widget()
            if type(child) == TreeItem:
                child.set_all_branches(hidden)

        self.children_widget.setHidden(hidden)
        self.label.setText(">" if self.children_widget.isHidden() else "v")


class ModifiedTagWidget(
    QWidget
):  # Needed to be modified because the original searched the display name in the library where it wasn't added yet
    def __init__(self, tag: Tag, parentTag: Tag) -> None:
        super().__init__()
        self.tag = tag

        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.base_layout = QVBoxLayout(self)
        self.base_layout.setObjectName("baseLayout")
        self.base_layout.setContentsMargins(0, 0, 0, 0)

        self.bg_button = QPushButton(self)
        self.bg_button.setFlat(True)
        if parentTag is not None:
            text = f"{tag.name} ({parentTag.name})".replace("&", "&&")
        else:
            text = tag.name.replace("&", "&&")
        self.bg_button.setText(text)
        self.bg_button.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)

        self.inner_layout = QHBoxLayout()
        self.inner_layout.setObjectName("innerLayout")
        self.inner_layout.setContentsMargins(2, 2, 2, 2)
        self.bg_button.setLayout(self.inner_layout)
        self.bg_button.setMinimumSize(math.ceil(22 * 1.5), 22)

        self.bg_button.setStyleSheet(
            f"QPushButton{{"
            f"background: {get_tag_color(ColorType.PRIMARY, tag.color)};"
            f"color: {get_tag_color(ColorType.TEXT, tag.color)};"
            f"font-weight: 600;"
            f"border-color:{get_tag_color(ColorType.BORDER, tag.color)};"
            f"border-radius: 6px;"
            f"border-style:inset;"
            f"border-width: {math.ceil(1*self.devicePixelRatio())}px;"
            f"padding-right: 4px;"
            f"padding-bottom: 1px;"
            f"padding-left: 4px;"
            f"font-size: 13px"
            f"}}"
            f"QPushButton::hover{{"
            f"border-color:{get_tag_color(ColorType.LIGHT_ACCENT, tag.color)};"
            f"}}"
        )

        self.base_layout.addWidget(self.bg_button)
        self.setMinimumSize(50, 20)



# -*- coding: utf-8 -*-

################################################################################
# Form generated from reading UI file 'home.ui'
##
# Created by: Qt User Interface Compiler version 6.5.1
##
# WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

# Copyright (C) 2024 Travis Abendshien (CyanVoxel).
# Licensed under the GPL-3.0 License.
# Created for TagStudio: https://github.com/CyanVoxel/TagStudio


import logging
import typing
from PySide6.QtCore import (QCoreApplication, QMetaObject, QRect,QSize, Qt)
from PySide6.QtGui import QFont
from PySide6.QtWidgets import (QComboBox, QFrame, QGridLayout,
                               QHBoxLayout, QVBoxLayout, QLayout, QLineEdit, QMainWindow,
                               QPushButton, QScrollArea, QSizePolicy,
                               QStatusBar, QWidget, QSplitter, QCheckBox,
                               QSpacerItem)
from src.qt.pagination import Pagination
from src.qt.widgets.landing import LandingWidget

# Only import for type checking/autocompletion, will not be imported at runtime.
if typing.TYPE_CHECKING:
    from src.qt.ts_qt import QtDriver

logging.basicConfig(format="%(message)s", level=logging.INFO)


class Ui_MainWindow(QMainWindow):
  
    def __init__(self, driver: "QtDriver", parent=None) -> None:
        super().__init__(parent)
        self.driver: "QtDriver" = driver
        self.setupUi(self)

        # NOTE: These are old attempts to allow for a translucent/acrylic
        # window effect. This may be attempted again in the future.
        # self.setWindowFlag(Qt.WindowType.NoDropShadowWindowHint, True)
        # self.setWindowFlag(Qt.WindowType.WindowTransparentForInput, False)
        # # self.setWindowFlag(Qt.WindowType.FramelessWindowHint, True)
        # self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

        # self.windowFX = WindowEffect()
        # self.windowFX.setAcrylicEffect(self.winId(), isEnableShadow=False)

        # # self.setStyleSheet(
        # # 	'background:#EE000000;'
        # # 	)
        

    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(1300, 720)
        
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.gridLayout = QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
  
        # ComboBox goup for search type and thumbnail size
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
  
        # left side spacer
        spacerItem = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
  
          # Search type selector
        self.comboBox_2 = QComboBox(self.centralwidget)
        self.comboBox_2.setMinimumSize(QSize(165, 0))
        self.comboBox_2.setObjectName("comboBox_2")
        self.comboBox_2.addItem("")
        self.comboBox_2.addItem("")
        self.horizontalLayout_3.addWidget(self.comboBox_2)
  
        # Thumbnail Size placeholder
        self.comboBox = QComboBox(self.centralwidget)
        self.comboBox.setObjectName(u"comboBox")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.comboBox.sizePolicy().hasHeightForWidth())
        self.comboBox.setSizePolicy(sizePolicy)
        self.comboBox.setMinimumWidth(128)
        self.comboBox.setMaximumWidth(128)
        self.horizontalLayout_3.addWidget(self.comboBox)
        self.gridLayout.addLayout(self.horizontalLayout_3, 5, 0, 1, 1)

        self.splitter = QSplitter()
        self.splitter.setObjectName(u"splitter")
        self.splitter.setHandleWidth(12)

        self.frame_container = QWidget()
        self.frame_layout = QVBoxLayout(self.frame_container)
        self.frame_layout.setSpacing(0)

        self.scrollArea = QScrollArea()
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setFocusPolicy(Qt.WheelFocus)
        self.scrollArea.setFrameShape(QFrame.NoFrame)
        self.scrollArea.setFrameShadow(QFrame.Plain)
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(
            u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 1260, 590))
        self.gridLayout_2 = QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setSpacing(8)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 8)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.frame_layout.addWidget(self.scrollArea)
        
        self.landing_widget: LandingWidget = LandingWidget(self.driver, self.devicePixelRatio())
        self.frame_layout.addWidget(self.landing_widget)

        self.pagination = Pagination()
        self.frame_layout.addWidget(self.pagination)

        self.horizontalLayout.addWidget(self.splitter)
        self.splitter.addWidget(self.frame_container)
        self.splitter.setStretchFactor(0, 1)

        self.gridLayout.addLayout(self.horizontalLayout, 10, 0, 1, 1)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setSizeConstraint(QLayout.SetMinimumSize)
        self.backButton = QPushButton(self.centralwidget)
        self.backButton.setObjectName(u"backButton")
        self.backButton.setMinimumSize(QSize(0, 32))
        self.backButton.setMaximumSize(QSize(32, 16777215))
        font = QFont()
        font.setPointSize(14)
        font.setBold(True)
        self.backButton.setFont(font)

        self.horizontalLayout_2.addWidget(self.backButton)

        self.forwardButton = QPushButton(self.centralwidget)
        self.forwardButton.setObjectName(u"forwardButton")
        self.forwardButton.setMinimumSize(QSize(0, 32))
        self.forwardButton.setMaximumSize(QSize(32, 16777215))
        font1 = QFont()
        font1.setPointSize(14)
        font1.setBold(True)
        font1.setKerning(True)
        self.forwardButton.setFont(font1)

        self.horizontalLayout_2.addWidget(self.forwardButton)

        self.searchField = QLineEdit(self.centralwidget)
        self.searchField.setObjectName(u"searchField")
        self.searchField.setMinimumSize(QSize(0, 32))
        font2 = QFont()
        font2.setPointSize(11)
        font2.setBold(False)
        self.searchField.setFont(font2)

        self.horizontalLayout_2.addWidget(self.searchField)

        self.searchButton = QPushButton(self.centralwidget)
        self.searchButton.setObjectName(u"searchButton")
        self.searchButton.setMinimumSize(QSize(0, 32))
        self.searchButton.setFont(font2)

        self.horizontalLayout_2.addWidget(self.searchButton)
        self.gridLayout.addLayout(self.horizontalLayout_2, 3, 0, 1, 1)
        self.gridLayout_2.setContentsMargins(6, 6, 6, 6)

        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(
            self.statusbar.sizePolicy().hasHeightForWidth())
        self.statusbar.setSizePolicy(sizePolicy1)
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate(
            "MainWindow", u"MainWindow", None))
        # Navigation buttons
        self.backButton.setText(
            QCoreApplication.translate("MainWindow", u"<", None))
        self.forwardButton.setText(
            QCoreApplication.translate("MainWindow", u">", None))
  
        # Search field
        self.searchField.setPlaceholderText(
            QCoreApplication.translate("MainWindow", u"Search Entries", None))
        self.searchButton.setText(
            QCoreApplication.translate("MainWindow", u"Search", None))
  
        # Search type selector
        self.comboBox_2.setItemText(0, QCoreApplication.translate("MainWindow", "And (Includes All Tags)"))
        self.comboBox_2.setItemText(1, QCoreApplication.translate("MainWindow", "Or (Includes Any Tag)"))
        self.comboBox.setCurrentText("")
  
        # Thumbnail size selector
        self.comboBox.setPlaceholderText(
            QCoreApplication.translate("MainWindow", u"Thumbnail Size", None))
    # retranslateUi

    def moveEvent(self, event) -> None:
        # time.sleep(0.02)  # sleep for 20ms
        pass

    def resizeEvent(self, event) -> None:
        # time.sleep(0.02)  # sleep for 20ms
        pass

    def toggle_landing_page(self, enabled: bool):
        if enabled:
            self.scrollArea.setHidden(True)
            self.landing_widget.setHidden(False)
            self.landing_widget.animate_logo_in()
        else:
            self.landing_widget.setHidden(True)
            self.landing_widget.set_status_label("")
            self.scrollArea.setHidden(False)